/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include "config_xor.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>

#include <errno.h>

/* Must be before <utime.h> on FreeBSD
 */
#include <sys/types.h>


#include <utime.h>


    

#ifdef HAVE_DIRENT_H
#include <dirent.h>
#define NAMLEN(dirent) sl_strlen((dirent)->d_name)
#else
#define dirent direct
#define NAMLEN(dirent) (dirent)->d_namlen
#ifdef HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif
#ifdef HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif
#ifdef HAVE_NDIR_H
#include <ndir.h>
#endif
#endif

#include "samhain.h"

#if (defined (SH_WITH_CLIENT) || defined (SH_STANDALONE)) 

#include "sh_error.h"
#include "sh_utils.h"
#include "sh_unix.h"
#include "sh_files.h"
#include "sh_tiger.h"
#include "sh_hash.h"

#undef  FIL__
#define FIL__  _("sh_files.c")


int sh_files_reportonce(char * c)
{
  int i;
  if (c == NULL)
    return (-1);

  i = sh_util_flagval(c);

  if      (i == S_TRUE)
    sh.flag.reportonce = S_TRUE;
  else if (i == S_FALSE)
    sh.flag.reportonce = S_FALSE;
  else
    return (-1);
  return 0;
}
    
int sh_files_fulldetail(char * c)
{
  int i;
  if (c == NULL)
    return (-1);

  i = sh_util_flagval(c);

  if      (i == S_TRUE)
    sh.flag.fulldetail = S_TRUE;
  else if (i == S_FALSE)
    sh.flag.fulldetail = S_FALSE;
  else
    return (-1);
  return 0;
}
    

typedef struct dir_struct {
  long    NumRegular;
  long    NumDirs;
  long    NumSymlinks;
  long    NumFifos;
  long    NumSockets;
  long    NumCDev;
  long    NumBDev;
  long    NumAll;
  long    TotalBytes;
  char    DirPath[PATH_MAX+2];
} dir_type;

typedef struct dirstack_entry {
  char                  * name;
  int                     class;
  unsigned long           check_mask;
  int                     rdepth;
  short                   checked;
  short                   childs_checked;
  short                   reported;
  struct dirstack_entry * next;
} dirstack_t;

static dirstack_t * dirList      = NULL;
static dirstack_t * fileList     = NULL;

static int        sh_files_fullpath  (char * testdir, char * d_name, 
				      char * statpath);
static int        sh_files_pushdir   (int class, char * str_s);
static int        sh_files_pushfile  (int class, char * str_s);
static int        sh_files_checkdir  (int class, int rdepth, char * dirName);
static ShFileType sh_files_filecheck (int class, char * dirName, 
				      char * fileName, long * total);

static long MaxRecursionLevel = 0;

/* set default recursion level
 */
int sh_files_setrecursion (char * flag_s)
{
  long flag = 0;
  static int reject = 0;
  if (reject == 1)
    return (-1);

  if (sh.flag.opts == 1)  
    reject = 1;

  if (flag_s != NULL) 
    flag = (int)(atof(flag_s));

  if (flag >= 0 && flag <= 99)
    MaxRecursionLevel = flag;
  else
    return (-1);

  return (0);
}


unsigned long sh_files_chk ()
{
  char       * tmp;
  ShFileType   status;

  dirstack_t * ptr;
  long         total;
  char       * base;
  char       * file;

  unsigned long fcount = 0;
  
  ptr = fileList;

  while (ptr) 
    {
      if (ptr->checked == S_FALSE)
	{
	  base = sh_util_basename (ptr->name);
	  file = sh_util_filename (ptr->name);
	  tmp = sh_util_safe_name (ptr->name);
	  
	  sh_error_handle ((-1),  FIL__, __LINE__, 0, MSG_FI_CHK,
			   tmp);
	  
	  status = sh_files_filecheck (ptr->class, base, file, &total);
	  

	  if (status == SH_FILE_UNKNOWN && ptr->reported == S_FALSE) 
	    {
	      if (sh.flag.checkSum == SH_CHECK_INIT || 
		  0 == sh_hash_have_it (ptr->name))
		  sh_error_handle ((ptr->class == SH_LEVEL_ALLIGNORE) ? 
				   ShDFLevel[ptr->class] : 
				   ShDFLevel[SH_ERR_T_FILE],
				   FIL__, __LINE__, 0, MSG_FI_FAIL,
				   tmp);
	      else
		  sh_error_handle (SH_ERR_INFO, FIL__, __LINE__, 0,
				   MSG_FI_FAIL,
				   tmp);
	      if (sh.flag.reportonce == S_TRUE)
		ptr->reported = S_TRUE;
	    }
	  else
	    ++fcount;
	  
	  SH_FREE(tmp);
	  SH_FREE(file);
	  SH_FREE(base);

	  ptr->checked = S_TRUE;
	}
      ptr = ptr->next;
    }

  return fcount;
}

int sh_files_delfilestack ()
{
  dirstack_t * old_ptr = fileList;

  while (old_ptr != NULL) 
    {
      fileList = old_ptr->next;
      SH_FREE (old_ptr->name);
      SH_FREE (old_ptr);
      old_ptr = fileList;
    }
  return (0);
}
  
int sh_files_deldirstack ()
{
  dirstack_t * old_ptr = dirList;

  while (old_ptr != NULL) {
    dirList = old_ptr->next;
    SH_FREE (old_ptr->name);
    SH_FREE (old_ptr);
    old_ptr = dirList;
  }
  return (0);
}

void sh_files_reset()
{
  dirstack_t * ptr = fileList;

  while (ptr != NULL) {
    ptr->checked = 0;
    ptr = ptr->next;
  }
 return;
}

void sh_dirs_reset()
{
  dirstack_t * ptr = dirList;

  while (ptr != NULL) {
    ptr->checked = 0;
    ptr = ptr->next;
  }
 return;
}


int sh_files_pushfile_ro (char * str_s)
{
  return (sh_files_pushfile (SH_LEVEL_READONLY, str_s));
}

int sh_files_pushfile_attr (char * str_s)
{
  return (sh_files_pushfile (SH_LEVEL_ATTRIBUTES, str_s));
}

int sh_files_pushfile_log (char * str_s)
{
  return (sh_files_pushfile (SH_LEVEL_LOGFILES, str_s));
}

int sh_files_pushfile_glog (char * str_s)
{
  return (sh_files_pushfile (SH_LEVEL_LOGGROW, str_s));
}

int sh_files_pushfile_noig (char * str_s)
{
  return (sh_files_pushfile (SH_LEVEL_NOIGNORE, str_s));
}

int sh_files_pushfile_allig (char * str_s)
{
  return (sh_files_pushfile (SH_LEVEL_ALLIGNORE, str_s));
}

static unsigned long sh_files_maskof (int class)
{
  switch (class)
    {
    case SH_LEVEL_ALLIGNORE:
      return (unsigned long) MASK_ALLIGNORE;
    case SH_LEVEL_NOIGNORE:
      return (unsigned long) MASK_NOIGNORE;
    case SH_LEVEL_ATTRIBUTES:
      return (unsigned long) MASK_ATTRIBUTES;
    case SH_LEVEL_LOGFILES:
      return (unsigned long) MASK_LOGFILES;
    case SH_LEVEL_LOGGROW:
      return (unsigned long) MASK_LOGGROW;
    case SH_LEVEL_READONLY:
      return (unsigned long) MASK_READONLY;
    default:
      return (unsigned long) 0;
    }
}


int sh_files_pushfile (int class, char * str_s)
{
  dirstack_t * new_item_ptr;
  char  * fileName;
  char  * tmp;
  int     len;

  static int reject = 0;

  if (reject == 1)
    return (-1);

  /* if we push a filename from the command line, make sure it
   * is the only one -- and will stay the only one
   */
  if (sh.flag.opts == 1) 
    {
      sh_files_delfilestack ();
      sh_files_deldirstack ();
      reject = 1;
    }

  if (str_s == NULL) 
    return (-1);

  len = sl_strlen(str_s);

  if (len > PATH_MAX) 
    {
      tmp = sh_util_safe_name (str_s);
      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_2LONG,
		       tmp);
      SH_FREE(tmp);
      return (-1);
    } 
  else if (len < 1) 
    {
      return (-1);
    } 
  else if (str_s[0] != '/') 
    {
      tmp = sh_util_safe_name (str_s);
      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_NOPATH,
		       tmp);
      SH_FREE(tmp);
      return (-1);
    } 
  else 
    {
      
      if (str_s[len-1] == '/' && len > 1)
	{
	  str_s[len-1] = '\0';
	  --len;
	}
      
      fileName = SH_ALLOC(len+1);
      sl_strlcpy(fileName, str_s, len+1);
    } 

  new_item_ptr = SH_ALLOC (sizeof(dirstack_t));

  new_item_ptr->name           = fileName;
  new_item_ptr->class          = class;
  new_item_ptr->check_mask     = sh_files_maskof(class);
  new_item_ptr->rdepth         = 0;
  new_item_ptr->checked        = S_FALSE;
  new_item_ptr->reported       = S_FALSE;
  new_item_ptr->childs_checked = S_FALSE;
  new_item_ptr->next           = fileList;

  fileList = new_item_ptr;
  return (0);
}


/* ------ directories ----- */

unsigned long sh_dirs_chk ()
{
  dirstack_t * ptr;
  int         status;
  unsigned long dcount = 0;
  char       * tmp;
  
  ptr = dirList;

  while (ptr) 
    {
      if (ptr->checked == S_FALSE)
	{
	  status = sh_files_checkdir (ptr->class, ptr->rdepth, ptr->name);
	  if (status < 0 && ptr->reported == S_FALSE) 
	    {
	      tmp = sh_util_safe_name (ptr->name);
	      sh_error_handle (ShDFLevel[SH_ERR_T_DIR], FIL__, __LINE__,
			       0, MSG_FI_FAIL, tmp);
	      SH_FREE(tmp);
	      if (sh.flag.reportonce == S_TRUE)
		ptr->reported = S_TRUE;
	    } 
	  else 
	    {
	      ++dcount;
	    }
	  ptr->checked = S_TRUE;
	}
      ptr = ptr->next;
    }
  return dcount;
}

int sh_files_pushdir_attr (char * str_s)
{
  return (sh_files_pushdir (SH_LEVEL_ATTRIBUTES, str_s));
}

int sh_files_pushdir_ro (char * str_s)
{
  return (sh_files_pushdir (SH_LEVEL_READONLY, str_s));
}

int sh_files_pushdir_log (char * str_s)
{
  return (sh_files_pushdir (SH_LEVEL_LOGFILES, str_s));
}

int sh_files_pushdir_glog (char * str_s)
{
  return (sh_files_pushdir (SH_LEVEL_LOGGROW, str_s));
}

int sh_files_pushdir_noig (char * str_s)
{
  return (sh_files_pushdir (SH_LEVEL_NOIGNORE, str_s));
}

int sh_files_pushdir_allig (char * str_s)
{
  return (sh_files_pushdir (SH_LEVEL_ALLIGNORE, str_s));
}


int sh_files_pushdir (int class, char * str_s)
{
  dirstack_t * new_item_ptr;
  char  * dirName;
  char  * tmp;
  int     len;
  int     rdepth = 0;
  char  * tail = NULL;

  if (sh.flag.opts == 1) {
    sh_files_delfilestack ();
    sh_files_deldirstack ();
  }

  if (str_s == NULL)
    return (-1);
  

  if (str_s[0] != '/')
    {
      rdepth = strtol(str_s, &tail, 10);
      if (tail == str_s)
	  return (-1);
    }
  else
    tail   = str_s;
  

  if (rdepth < (-1) || tail == str_s || rdepth > 99)
    rdepth = MaxRecursionLevel;

  len = sl_strlen(tail);

  if (len > PATH_MAX) 
    {
      tmp = sh_util_safe_name (tail);
      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_2LONG,
		       tmp);
      SH_FREE(tmp);
      return (-1);
    } 
  else if (len < 1) 
    {
      return (-1);
    } 
  else if (tail[0] != '/') 
    {
      tmp = sh_util_safe_name (tail);
      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_NOPATH,
		       tmp);
      SH_FREE(tmp);
      return (-1);
    } 
  else 
    {
      
      if (tail[len-1] == '/' && len > 1)
	{
	  tail[len-1] = '\0';
	  --len;
	}
      
      dirName = SH_ALLOC(len+1);
      sl_strlcpy(dirName, tail, len+1);
    } 
  
  new_item_ptr = SH_ALLOC (sizeof(dirstack_t));

  new_item_ptr->name           = dirName;
  new_item_ptr->class          = class;
  new_item_ptr->check_mask     = sh_files_maskof(class);
  new_item_ptr->rdepth         = rdepth;
  new_item_ptr->checked        = S_FALSE;
  new_item_ptr->reported       = S_FALSE;
  new_item_ptr->childs_checked = S_FALSE;
  new_item_ptr->next           = dirList;


  dirList = new_item_ptr;
  return (0);
}
  


int sh_files_checkdir (int iclass, int idepth, char * iname)
{
  DIR *           thisDir = NULL;
  struct dirent * thisEntry;
  int             status;
  dir_type        theDir;
  ShFileType      checkit;


  file_type       theFile;
  char          * tmpname;
  char          * tmpcat;

  int             rdepth = 0;
  int             class  = 0;
  int             rdepth_next;
  int             class_next;

  int             checked_flag  = S_FALSE;
  int             cchecked_flag = S_FALSE;

  dirstack_t *    dst_ptr;


  if (iname == NULL || idepth < (-1))
    return (-1);
  
  if (idepth < 0)
    return (0);
  
  rdepth = idepth;
  class  = iclass;
  
  tmpname = sh_util_safe_name (iname);

  /* ---- check for obscure name ----
   */
  sh_util_obscurename (ShDFLevel[SH_ERR_T_NAME], iname);
  
  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_CHK,
		   tmpname);
  
  /* ---- check input ----
   */
  if ( sl_strlen(iname) > PATH_MAX) 
    {
      sh_error_handle (ShDFLevel[SH_ERR_T_DIR], FIL__, __LINE__, 0, 
		       MSG_FI_2LONG,
		       tmpname);
      SH_FREE(tmpname);
      return (-1);
    }
  
  /* ---- check for absolute path ---- */
  if ( iname[0] != '/') 
    {
      sh_error_handle (ShDFLevel[SH_ERR_T_DIR], FIL__, __LINE__, 0, 
		       MSG_FI_NOPATH,
		       tmpname);
      SH_FREE(tmpname);
      return (-1);
    }
    
    
  /* ---- stat the directory ----
   */
  sl_strlcpy (theFile.fullpath, iname, PATH_MAX+1);
  status = sh_unix_getinfo (ShDFLevel[SH_ERR_T_DIR], 
			    iname,
			    &theFile);
  if (theFile.c_mode[0] != 'd') 
    { 
      sh_error_handle (ShDFLevel[SH_ERR_T_DIR], FIL__, __LINE__, 0,
		       MSG_FI_NODIR,
		       tmpname);
      SH_FREE(tmpname); 
      return (-1);
    }

  /* ---- open directory for reading ---- 
   */
  sl_set_suid();
  thisDir = opendir (iname);
  sl_unset_suid();

  if (thisDir == NULL) 
    {
      status = errno;
      sh_error_handle (ShDFLevel[SH_ERR_T_DIR], FIL__, __LINE__, 0, 
		       MSG_E_OPENDIR,
		       sh_error_message (status), tmpname);
      SH_FREE(tmpname); 
      return (-1);
    }

  theDir.NumRegular  = 0;
  theDir.NumDirs     = 0;
  theDir.NumSymlinks = 0;
  theDir.NumFifos    = 0;
  theDir.NumSockets  = 0;
  theDir.NumCDev     = 0;
  theDir.NumBDev     = 0;
  theDir.NumAll      = 0;
  theDir.TotalBytes  = 0;
  sl_strlcpy (theDir.DirPath, iname, PATH_MAX+1); 

  /* ---- read ----
   */
  do {
    thisEntry = readdir (thisDir);

    if (thisEntry != NULL) {
      ++theDir.NumAll;

      if (sl_strcmp (thisEntry->d_name, ".") == 0)
	{ 
	  ++theDir.NumDirs;
	  continue;
	}

      if (sl_strcmp (thisEntry->d_name, "..") == 0)
	{
	  ++theDir.NumDirs;
	  continue;
	}

      /* ---- Check the file. ---- 
       */
      tmpcat = SH_ALLOC(PATH_MAX+2);
      sl_strlcpy(tmpcat, iname,                 PATH_MAX+2);
      sl_strlcat(tmpcat, "/",                   PATH_MAX+2);
      sl_strlcat(tmpcat, thisEntry->d_name,     PATH_MAX+2);

      dst_ptr       = fileList;
      rdepth_next   = rdepth - 1;
      class_next    = class;
      checked_flag  = -1;
      cchecked_flag = -1;

      while (dst_ptr) 
	{
	  if (0 == sl_strcmp(dst_ptr->name,  tmpcat))
	    {
	      class_next    = dst_ptr->class;
	      checked_flag  = dst_ptr->checked;
	      cchecked_flag = dst_ptr->childs_checked;
	      break;
	    }
	  dst_ptr = dst_ptr->next;
	}

      /* ---- Has been checked already. ----
       */
      if (checked_flag == S_TRUE && cchecked_flag == S_TRUE)
	continue;
	  
      /* --- May be true, false, or not found. --- 
       */
      if (checked_flag == S_TRUE)
	{
	  /* -- need only the file type --
	   */
	  checkit = sh_unix_get_ftype(tmpcat);
	}
      else
	{
	  /* -- need to check the file itself --
	   */
	  checkit = sh_files_filecheck (class_next, 
					iname, 
					thisEntry->d_name,
					&theDir.TotalBytes);
	  if (checked_flag == S_FALSE)
	    dst_ptr->checked = S_TRUE;
	}
	  
      if      (checkit == SH_FILE_REGULAR)   
	++theDir.NumRegular;

      else if (checkit == SH_FILE_DIRECTORY) 
	{
	  ++theDir.NumDirs;
	  if (rdepth_next >= 0 && cchecked_flag != S_TRUE) 
	    {
	      rdepth_next = rdepth - 1;

	      /* check whether the new directory is in the
	       * list with a recursion depth already defined
	       */
	      checked_flag  = -1;
	      cchecked_flag = -1;

	      dst_ptr     = dirList;
	      while (dst_ptr) 
		{
		  if (0 == sl_strcmp(dst_ptr->name, tmpcat))
		    {
		      rdepth_next   = dst_ptr->rdepth;
		      class_next    = dst_ptr->class;
		      checked_flag  = dst_ptr->childs_checked;
		      cchecked_flag = dst_ptr->checked;
		      break;
		    }
		  dst_ptr = dst_ptr->next;
		}

	      if (cchecked_flag == S_FALSE)
		{
		  sh_files_checkdir (class_next, rdepth_next, tmpcat);
		  dst_ptr->childs_checked = S_TRUE;
		}
	      else if (checked_flag == -1)
		sh_files_checkdir (class_next, rdepth_next, tmpcat);
		    
	    }
	}

      else if (checkit == SH_FILE_SYMLINK)   ++theDir.NumSymlinks;
      else if (checkit == SH_FILE_FIFO)      ++theDir.NumFifos;
      else if (checkit == SH_FILE_SOCKET)    ++theDir.NumSockets;
      else if (checkit == SH_FILE_CDEV)      ++theDir.NumCDev;
      else if (checkit == SH_FILE_BDEV)      ++theDir.NumBDev;

      SH_FREE(tmpcat);

    }
  } while (thisEntry != NULL);


  sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_DSUM,
		   theDir.NumDirs,
		   theDir.NumRegular,
		   theDir.NumSymlinks,
		   theDir.NumFifos,
		   theDir.NumSockets,
		   theDir.NumCDev,
		   theDir.NumBDev);

  
  closedir (thisDir);
  SH_FREE(tmpname);
  return (0);
}

ShFileType sh_files_filecheck (int class, char * dirName, char * fileName,
			       long * total)
{
  char            fullpath[PATH_MAX+2];
  char            fileHash[KEY_LEN + 1];
  int             status;
  file_type       theFile;
  char          * tmpdir;
  char          * tmpname;
  struct utimbuf  utime_buf;

  if (dirName == NULL || fileName == NULL)
    {
     
      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_NULL);
      return SH_FILE_UNKNOWN;
    }

  if (0 != sh_util_obscurename (ShDFLevel[SH_ERR_T_NAME], 
				  fileName)) 
    {
      tmpdir  = sh_util_safe_name (dirName);
      tmpname = sh_util_safe_name (fileName);
      
      sh_error_handle (ShDFLevel[SH_ERR_T_NAME], FIL__, __LINE__, 0,
		       MSG_FI_OBSC2,
		       tmpdir, tmpname);
      SH_FREE(tmpname);
      SH_FREE(tmpdir);
    }    

  if (0 != sh_files_fullpath (dirName, fileName, fullpath)) 
    { 
      tmpdir  = sh_util_safe_name (dirName);
      tmpname = sh_util_safe_name (fileName);
      sh_error_handle (ShDFLevel[SH_ERR_T_FILE],  FIL__, __LINE__, 0,
		       MSG_FI_2LONG2,
		       tmpdir, tmpname);
      SH_FREE(tmpname);
      SH_FREE(tmpdir);
      return SH_FILE_UNKNOWN; 
    } 


  /* stat the file
   */
  sl_strlcpy (theFile.fullpath, fullpath, PATH_MAX+1);
  theFile.check_mask = sh_files_maskof(class);
  theFile.reported = S_FALSE;

  status = sh_unix_getinfo ( (class == SH_LEVEL_ALLIGNORE) ? 
			     ShDFLevel[class] : ShDFLevel[SH_ERR_T_FILE], 
			     fileName,
			     &theFile);
  
  if (status != 0)
    {
      if (class == SH_LEVEL_ALLIGNORE)
	  sh_hash_set_visited (fullpath);
      return SH_FILE_UNKNOWN;
    }
  
  /* compute checksum 
   *
   * 1 == init
   * 2 == check
   */
  if (theFile.c_mode[0] == '-')
    {
      sl_strlcpy(fileHash, 
		 sh_tiger_hash (fullpath, TIGER_FILE, 0), 
		 KEY_LEN + 1);
      sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_CSUM,
		       fileHash, fullpath);
    } 
  else
    {
      sl_strlcpy(fileHash, 
		 _("000000000000000000000000000000000000000000000000"), 
		 KEY_LEN + 1);
    }
      
  if (sh.flag.checkSum == SH_CHECK_INIT) 
    {
      sh_hash_pushdata (&theFile, fileHash);
    } 
  else if (sh.flag.checkSum == SH_CHECK_CHECK 
	   /* && theFile.c_mode[0] == '-' */
	   /* && class != SH_LEVEL_ALLIGNORE */
	   ) 
    {
      sh_hash_compdata (class, &theFile, fileHash);
    }
  
  
  /* reset the access time 
   */
  if (class == SH_LEVEL_NOIGNORE)
    {
      utime_buf.actime   = (time_t) theFile.atime;
      utime_buf.modtime  = (time_t) theFile.mtime;
#if !defined(O_NOATIME)
      sl_set_suid ();
      retry_aud_utime (FIL__, __LINE__, fullpath, &utime_buf);
      sl_unset_suid ();
#endif
    }
  
  /* yes I know, the 'break' is not needed
   */
  switch (theFile.c_mode[0]) 
    {
    case '-': return SH_FILE_REGULAR;   break;
    case 'l': return SH_FILE_SYMLINK;   break;
    case 'd': return SH_FILE_DIRECTORY; break;
    case 'c': return SH_FILE_CDEV;      break;
    case 'b': return SH_FILE_BDEV;      break;
    case '|': return SH_FILE_FIFO;      break;
    case 's': return SH_FILE_SOCKET;    break;
    default:  return SH_FILE_UNKNOWN;   break;
    }
  
  /* notreached */
}


static int sh_files_fullpath (char * testdir, char * d_name, char * statpath)
{
  if (testdir != NULL) 
    {
      if (sl_strlen(testdir) > (PATH_MAX-2) ) 
	return (-1);
      sl_strlcpy(statpath, testdir,    PATH_MAX - 2);
    }
  if (d_name != NULL) 
    {
      sl_strlcat(statpath, "/",   PATH_MAX+1);
      if (sl_strlen(d_name) >= ( PATH_MAX - sl_strlen(statpath) ) ) 
	return (-1);
      sl_strlcat(statpath, d_name,   PATH_MAX+1);
    }
  if (statpath == NULL) 
    return (-1);
  return (0);
}


/* -----------------------------------
 * 
 *  The following two routines serve to
 *  verify that the user has selected
 *  a proper setup for file policies.
 *
 * -----------------------------------
 */
static int check_file(char * name)
{
  dirstack_t * pfilL = fileList;

  while (pfilL)
    {
      if (0 == strcmp(name, pfilL->name) &&
	  (pfilL->check_mask & MODI_ATM) == 0 &&
	  (pfilL->check_mask & MODI_CTM) == 0 &&
	  (pfilL->check_mask & MODI_MTM) == 0)
	return 0;
      pfilL = pfilL->next;
    }
  return (-1);
}
  
int sh_files_test_setup ()
{
  int dlen, flen;

  dirstack_t * pdirL = dirList;  
  dirstack_t * pfilL = fileList;

  while (pdirL)
    {
      dlen = strlen(pdirL->name);
      pfilL = fileList;
      while (pfilL)
	{
	  flen = strlen(pfilL->name);

	  /* check whether file is in tree of dir
	   */
	  if ((pfilL->class == SH_LEVEL_READONLY) ||
	      (pfilL->class == SH_LEVEL_NOIGNORE))
	    {
	      ;  /* do nothing */
	    }
	  else
	    {
	      if ((flen > dlen) && 
		  (pfilL->name[dlen] == '/') &&
		  (0 == strncmp(pfilL->name, pdirL->name, dlen)))
		{
		  if ((pdirL->check_mask & MODI_ATM) != 0  ||
		      (pdirL->check_mask & MODI_MTM) != 0  ||
		      (pdirL->check_mask & MODI_CTM) != 0)
		    {
		      if (check_file (pdirL->name) != 0)
			sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_COLL,
					 pdirL->name, pfilL->name);
		    }
		}
	    }
	  
	  pfilL = pfilL->next;
	}


      pfilL = dirList;
      while (pfilL)
	{
	  flen = strlen(pfilL->name);

	  /* check whether file is in tree of dir
	   */
	  if ((pfilL->class == SH_LEVEL_READONLY) ||
	      (pfilL->class == SH_LEVEL_NOIGNORE))
	    {
	      ;  /* do nothing */
	    }
	  else
	    {
	      if ((flen > dlen) && 
		  (pfilL->name[dlen] == '/') &&
		  (0 == strncmp(pfilL->name, pdirL->name, dlen)))
		{
		  if ((pdirL->check_mask & MODI_ATM) != 0  ||
		      (pdirL->check_mask & MODI_MTM) != 0  ||
		      (pdirL->check_mask & MODI_CTM) != 0)
		    {
		      if (check_file (pdirL->name) != 0)
			sh_error_handle ((-1), FIL__, __LINE__, 0, MSG_FI_COLL,
					 pdirL->name, pfilL->name);
		    }
		}
	    }
	  
	  pfilL = pfilL->next;
	}

      pdirL = pdirL->next;
    }

  return 0;
}
      


#endif
